# w.c.s. - web application for online forms
# Copyright (C) 2005-2022  Entr'ouvert
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, see <http://www.gnu.org/licenses/>.

import io
from contextlib import contextmanager

from django.core.handlers.wsgi import WSGIRequest
from quixote import get_publisher

from wcs import sql
from wcs.compat import CompatHTTPRequest
from wcs.fields import Field, PageField
from wcs.qommon.form import FileWithPreviewWidget, Form, get_selection_error_text
from wcs.qommon.template import TemplateError
from wcs.sql_criterias import Equal
from wcs.workflows import WorkflowStatusItem

from .qommon import _, misc


class TestError(Exception):
    def __init__(self, msg, error=None):
        self.msg = msg
        self.error = error or msg

    # prevent pytest from trying to collect this class (#75521)
    __test__ = False


class TestDef(sql.TestDef):
    _names = 'testdef'

    name = ''
    slug = None
    object_type = None  # (formdef, carddef, etc.)
    object_id = None

    data = None  # (json export of formdata, carddata, etc.)
    is_in_backoffice = False
    expected_error = None

    ignored_field_types = (
        'subtitle',
        'title',
        'comment',
        'computed',
        'table',
        'table-select',
        'tablerows',
        'ranked-items',
    )

    def __str__(self):
        return self.name

    def get_admin_url(self):
        base_url = get_publisher().get_backoffice_url()
        return '%s/forms/%s/tests/%s/' % (base_url, self.object_id, self.id)

    def store(self, *args, comment=None, snapshot_store_user=True, **kwargs):
        if not self.slug:
            existing_slugs = {
                x.slug
                for x in self.select(
                    [Equal('object_type', self.object_type), Equal('object_id', self.object_id)]
                )
            }
            base_slug = misc.simplify(self.name)
            self.slug = base_slug
            i = 2
            while self.slug in existing_slugs:
                self.slug = '%s-%s' % (base_slug, i)
                i += 1
        super().store(*args, **kwargs)

    @classmethod
    def select_for_objectdef(cls, objectdef):
        return cls.select(
            [Equal('object_type', objectdef.get_table_name()), Equal('object_id', objectdef.id)]
        )

    @classmethod
    def create_from_formdata(cls, formdef, formdata):
        testdef = cls()
        testdef.object_type = formdef.get_table_name()
        testdef.object_id = formdef.id
        testdef.is_in_backoffice = formdata.backoffice_submission

        field_data = {}
        for field in formdef.fields:
            if field.id in formdata.data:
                value = formdata.data[field.id]

                if value and hasattr(field, 'get_json_value'):
                    value = field.get_json_value(value)

                field_data[field.id] = value

            for suffix in ('raw', 'display', 'structured'):
                key = '%s_%s' % (field.id, suffix)
                if key in formdata.data:
                    field_data[key] = formdata.data[key]

        testdef.data = {
            'fields': field_data,
            'user': formdata.user.get_json_export_dict() if formdata.user else None,
        }
        return testdef

    def build_formdata(self, objectdef, include_fields=False):
        formdata = objectdef.data_class()()
        if self.data['user']:
            formdata.set_user_from_json(self.data['user'])

        if include_fields:
            for field in objectdef.fields:
                if field.id not in self.data['fields']:
                    continue

                value = self.data['fields'].get(field.id)
                if value is not None:
                    value = field.from_json_value(value)

                self.add_value_to_formdata(field, formdata, value)

        return formdata

    @contextmanager
    def fake_request(self):
        def record_error(self, error_summary=None, context=None, exception=None, *args, **kwargs):
            raise exception

        real_record_error = get_publisher().record_error

        true_request = get_publisher().get_request()
        wsgi_request = WSGIRequest({'REQUEST_METHOD': 'POST', 'wsgi.input': io.StringIO()})
        fake_request = CompatHTTPRequest(wsgi_request)
        fake_request.is_in_backoffice_forced_value = self.is_in_backoffice
        try:
            get_publisher()._set_request(fake_request)
            get_publisher().record_error = record_error
            yield
        finally:
            get_publisher()._set_request(true_request)
            get_publisher().record_error = real_record_error

    def run(self, objectdef):
        self.missing_required_fields = []
        with self.fake_request():
            try:
                self._run(objectdef)
            except TestError as e:
                if not self.expected_error:
                    raise e

                if e.error != self.expected_error:
                    raise TestError(
                        _('Expected error "%(expected_error)s" but got error "%(error)s" instead.')
                        % {'expected_error': self.expected_error, 'error': e.error}
                    )
            else:
                if self.expected_error:
                    raise TestError(
                        _('Expected error "%s" but test completed with success.') % self.expected_error
                    )

    def _run(self, objectdef):
        formdata = self.build_formdata(objectdef)

        get_publisher().reset_formdata_state()
        get_publisher().substitutions.feed(objectdef)
        get_publisher().substitutions.feed(formdata)

        self.form = Form(action='#')

        fields = []
        fields_by_page = {}
        for field in objectdef.fields:
            if field.key == 'page':
                fields = fields_by_page[field] = []
                continue
            fields.append(field)

        if not fields_by_page:  # form without pages
            fields_by_page[PageField()] = fields

        previous_page = None
        for i, (page, fields) in enumerate(fields_by_page.items(), 1):
            page.index = i

            if previous_page:
                self.evaluate_page_conditions(previous_page, formdata, objectdef)

            if page and not page.is_visible(formdata.data, objectdef):
                fields_with_data = [
                    field for field in fields if self.data['fields'].get(field.id) is not None
                ]
                if fields_with_data:
                    raise TestError(
                        _('Tried to fill field "%(label)s" on page %(no)d but page was not shown.')
                        % {'label': fields_with_data[0].label, 'no': page.index}
                    )
                continue

            self.fill_page_fields(fields, page, formdata, objectdef)
            previous_page = page

        if previous_page:  # evaluate last page post conditions
            self.evaluate_page_conditions(previous_page, formdata, objectdef)

    def fill_page_fields(self, fields, page, formdata, objectdef):
        self.handle_computed_fields(fields, formdata)
        for field in fields:
            if field.key in self.ignored_field_types:
                continue

            if not field.is_visible(formdata.data, objectdef):
                if self.data['fields'].get(field.id) is not None:
                    raise TestError(
                        _('Tried to fill field "%(label)s" on page %(no)d but it is hidden.')
                        % {'label': field.label, 'no': page.index}
                    )
                continue

            # make sure to never request data source
            if hasattr(field, 'data_source'):
                field._data_source = field.data_source
                field.data_source = None
            elif hasattr(field, 'block'):
                for x in field.block.fields:
                    x.data_source = None

            value = self.data['fields'].get(field.id)
            if value is not None:
                value = field.from_json_value(value)

            self.run_widget_validation(field, value)

            self.add_value_to_formdata(field, formdata, value)

            get_publisher().substitutions.invalidate_cache()

        self.handle_computed_fields(fields, formdata, exclude_frozen=True)

    def add_value_to_formdata(self, field, formdata, value):
        if field.key in ('item', 'items') and getattr(field, '_data_source', field.data_source):
            # add values without requesting data source
            formdata.data[field.id] = value
            for suffix in ('raw', 'display', 'structured'):
                key = '%s_%s' % (field.id, suffix)
                if key in self.data['fields']:
                    formdata.data[key] = self.data['fields'][key]
        else:
            field.set_value(formdata.data, value)

    def evaluate_page_conditions(self, page, formdata, objectdef):
        for post_condition in page.post_conditions or []:
            condition = post_condition.get('condition', {})
            try:
                if not Field.evaluate_condition(formdata.data, objectdef, condition, record_errors=False):
                    raise TestError(
                        _('Page %(no)d post condition was not met (%(condition)s).')
                        % {'no': page.index, 'condition': condition.get('value')},
                        error=post_condition.get('error_message'),
                    )
            except RuntimeError:
                raise TestError(_('Failed to evaluate page %d post condition.') % page.index)

    def run_widget_validation(self, field, value):
        widget = field.add_to_form(self.form)

        if isinstance(widget, FileWithPreviewWidget):
            widget.get_value_from_token = False

        widget.set_value(value)
        widget.transfer_form_value(get_publisher().get_request())

        widget._parsed = False
        widget.parse()

        widget = TestDef.get_error_widget(widget, self)
        if not widget:
            return

        field_label = _('"%s"') % field.label

        if getattr(widget, 'is_subwidget', False):
            value = widget.value
            field = widget.field
            field_label = _('"%(subfield)s" (of field %(field)s)') % {
                'subfield': field.label,
                'field': field_label,
            }

        if field.convert_value_to_str:
            value = field.convert_value_to_str(value)

        error_msg = _('Invalid value "%s"') % value if value else _('Empty value')
        raise TestError(
            _('%(error)s for field %(label)s: %(details)s.')
            % {
                'error': error_msg,
                'label': field_label,
                'details': widget.error,
            },
            error=widget.error,
        )

    def handle_computed_fields(self, fields, formdata, exclude_frozen=False):
        for field in fields:
            if field.key != 'computed':
                continue
            if exclude_frozen and field.freeze_on_initial_value:
                continue

            with get_publisher().complex_data():
                try:
                    value = WorkflowStatusItem.compute(field.value_template, raises=True, allow_complex=True)
                except TemplateError:
                    continue
                else:
                    value = get_publisher().get_cached_complex_data(value)

                if isinstance(value, str) and len(value) > 10000:
                    value = None

                formdata.data[field.id] = value
                get_publisher().substitutions.invalidate_cache()

    @staticmethod
    def widget_has_real_error(widget, testdef):
        if widget.error == widget.REQUIRED_ERROR:
            if testdef:
                label = widget.block.name if hasattr(widget, 'block') else widget.field.label
                testdef.missing_required_fields.append(label)
            return False

        if widget.error != get_selection_error_text():
            return True

        return False

    @classmethod
    def get_error_widget(cls, widget, testdef=None):
        if not widget.has_error():
            return

        if widget.field.key == 'block' and (not widget.error or widget.error == widget.REQUIRED_ERROR):
            widget.error = None
            return cls.get_error_subwidget(widget, testdef)

        if cls.widget_has_real_error(widget, testdef):
            return widget

    @classmethod
    def get_error_subwidget(cls, widget, testdef):
        for widget in widget.get_widgets():
            widget.is_subwidget = True

            if widget.error and cls.widget_has_real_error(widget, testdef):
                return widget

            if hasattr(widget, 'get_widgets'):
                widget = TestDef.get_error_subwidget(widget, testdef)
                if widget:
                    return widget

    def export_to_json(self):
        return {
            'name': self.name,
            'slug': self.slug,
            'object_type': self.object_type,
            'object_id': self.object_id,
            'data': self.data,
        }

    @classmethod
    def import_from_json(cls, data):
        testdefs = TestDef.select(
            [
                Equal('object_type', data['object_type']),
                Equal('object_id', data['object_id']),
                Equal('slug', data['slug']),
            ]
        )

        testdef = testdefs[0] if testdefs else TestDef()

        for k, v in data.items():
            setattr(testdef, k, v)

        testdef.store()
        return testdef


class TestResult(sql.TestResult):
    _names = 'test_result'

    object_type = None  # (formdef, carddef, etc.)
    object_id = None
    timestamp = None
    success = None
    reason = None  # reason for tests execution
    results = None  # results for each test associated to object

    def get_admin_url(self):
        base_url = get_publisher().get_backoffice_url()
        return '%s/forms/%s/tests/results/%s/' % (base_url, self.object_id, self.id)
